/**
 * @file Virtual 8-Bit Computer
 * @copyright Alexander Bass 2024
 * @license GPL-3.0
 */
import { Computer } from "./computer";
import { $ } from "./etc";
import { ISA } from "./instructionSet";
import { generate_isa } from "./isaGenerator";
import { UI } from "./ui";
import { u8 } from "./num";

import "./style/style.scss";
import { CpuEvent } from "./events";

declare global {
	interface Window {
		comp: Computer;
		ui: UI;
		firehose: () => void;
	}
}

function main(): void {
	const program: Array<u8> = [
		0x19, 0x00, 0xf0, 0x14, 0x00, 0x01, 0x5e, 0x00, 0xf0, 0x01, 0x21, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x48, 0x65, 0x6c, 0x6c, 0x6f, 0x20, 0x57,
		0x6f, 0x72, 0x6c, 0x64, 0x21, 0x0a, 0x00, 0x00, 0x00,
	];
	const computer = new Computer();

	const ui = new UI();
	ui.init_events(computer.events);
	computer.load_memory(program);
	computer.init_events(ui.events);
	window.comp = computer;
	window.ui = ui;

	$("ISA").textContent = generate_isa(ISA);

	$("binary_upload").addEventListener("change", (e) => {
		const t = e.target;
		if (t === null) {
			return;
		}

		const file: File | undefined = (t as HTMLInputElement).files?.[0];
		if (file === undefined) {
			console.log("No files attribute on file input");
			return;
		}
		const reader = new FileReader();
		console.log(file);
		reader.addEventListener("load", (e) => {
			if (e.target !== null) {
				const data = e.target.result;
				if (data instanceof ArrayBuffer) {
					const view = new Uint8Array(data);
					const array = [...view] as Array<u8>;
					computer.reset();
					computer.load_memory(array);
				} else {
					console.log("not array");
				}
			}
		});
		reader.readAsArrayBuffer(file);
	});
	let fire = false;
	window.firehose = (): void => {
		if (fire === false) {
			computer.events.firehose((ident, data) => {
				console.log(`New Event: ${CpuEvent[ident]}. data: `, data);
			});
			fire = true;
		} else {
			console.error("Firehose already started");
		}
	};

	$("save_button").addEventListener("click", () => {
		const memory = computer.dump_memory();
		const blob = new Blob([memory], { type: "application/octet-stream" });
		const url = URL.createObjectURL(blob);

		const link = document.createElement("a");
		link.href = url;
		link.download = "bin.bin";
		link.click();
	});
}

document.addEventListener("DOMContentLoaded", () => {
	main();
});
